{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "\n# Simple UNet model.\n\nThis model is a simplified version of the U-Net architecture, \nwhich is widely used for image segmentation tasks.  \nThis is implemented in the proprietary FASTMRI package [fastmri]_.  \n\nThe U-Net model consists of an encoder (downsampling path) and \na decoder (upsampling path) with skip connections between corresponding \nlayers in the encoder and decoder. \nThese skip connections help in retaining spatial information \nthat is lost during the downsampling process.\n\nThe primary purpose of this model is to perform image reconstruction tasks, \nspecifically for MRI images. \nIt takes an input MRI image and reconstructs it to improve the image quality \nor to recover missing parts of the image.\n\nThis implementation of the UNet model was pulled from the FastMRI Facebook \nrepository, which is a collaborative research project aimed at advancing \nthe field of medical imaging using machine learning techniques.\n\n\\begin{align}\\mathbf{\\hat{x}} = \\mathrm{arg} \\min_{\\mathbf{x}} || \\mathcal{U}_\\mathbf{\\theta}(\\mathbf{y}) - \\mathbf{x} ||_2^2\\end{align}\n\nwhere:\n- \\( \\mathbf{\\hat{x}} \\) is the reconstructed MRI image,\n- \\( \\mathbf{x} \\) is the ground truth image,\n- \\( \\mathbf{y} \\) is the input MRI image (e.g., k-space data),\n- \\( \\mathcal{U}_\\mathbf{\\theta} \\) is the U-Net model parameterized by \\( \\theta \\).\n\n<div class=\"alert alert-danger\"><h4>Warning</h4><p>We train on a single image here. In practice, this should be done on a database like fastMRI [fastmri]_.</p></div>\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Imports\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "import os\nfrom pathlib import Path\nimport shutil\nimport brainweb_dl as bwdl\nimport matplotlib.pyplot as plt\nimport numpy as np\nimport torch\nfrom tqdm import tqdm\nimport time\nimport joblib\nfrom PIL import Image\nimport tempfile as tmp\n\nfrom fastmri.models import Unet\nfrom mrinufft import get_operator\nfrom mrinufft.trajectories import initialize_2D_cones"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Setup a simple class for the U-Net model\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "class Model(torch.nn.Module):\n    \"\"\"Model for MRI reconstruction using a U-Net.\"\"\"\n\n    def __init__(self, initial_trajectory):\n        super().__init__()\n        self.operator = get_operator(\"gpunufft\", wrt_data=True)(\n            initial_trajectory,\n            shape=(256, 256),\n            density=True,\n            squeeze_dims=False,\n        )\n        self.unet = Unet(in_chans=1, out_chans=1, chans=32, num_pool_layers=4)\n\n    def forward(self, kspace):\n        \"\"\"Forward pass of the model.\"\"\"\n        image = self.operator.adj_op(kspace)\n        recon = self.unet(image.float()).abs()\n        recon /= torch.mean(recon)\n        return recon"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Utility function to plot the state of the model\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "def plot_state(axs, mri_2D, traj, recon, loss=None, save_name=None):\n    \"\"\"Image plotting function.\n\n    Plot the original MRI image, the trajectory, the reconstructed image,\n    and the loss curve (if provided). Saves the plot if a filename is provided.\n\n    Parameters\n    ----------\n    axs (numpy array): Array of matplotlib axes to plot on.\n    mri_2D (torch.Tensor): Original MRI image.\n    traj : Trajectory.\n    recon (torch.Tensor): Reconstructed image after training.\n    loss (list, optional): List of loss values to plot. Defaults to None.\n    save_name (str, optional): Filename to save the plot. Defaults to None.\n    \"\"\"\n    axs = axs.flatten()\n    axs[0].imshow(np.abs(mri_2D[0]), cmap=\"gray\")\n    axs[0].axis(\"off\")\n    axs[0].set_title(\"MR Image\")\n    axs[1].scatter(*traj.T, s=0.5)\n    axs[1].set_title(\"Trajectory\")\n    axs[2].imshow(np.abs(recon[0][0].detach().cpu().numpy()), cmap=\"gray\")\n    axs[2].axis(\"off\")\n    axs[2].set_title(\"Reconstruction\")\n    if loss is not None:\n        axs[3].plot(loss)\n        axs[3].grid(\"on\")\n        axs[3].set_title(\"Loss\")\n    if save_name is not None:\n        plt.savefig(save_name, bbox_inches=\"tight\")\n        plt.close()\n    else:\n        plt.show()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Setup Inputs (models, trajectory and image)\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "init_traj = initialize_2D_cones(32, 256).reshape(-1, 2).astype(np.float32)\nmodel = Model(init_traj)\nmodel.eval()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Get the image on which we will train our U-Net Model\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "mri_2D = torch.Tensor(np.flipud(bwdl.get_mri(4, \"T1\")[80, ...]).astype(np.complex64))[\n    None\n]\nmri_2D = mri_2D / torch.mean(mri_2D)\nkspace_mri_2D = model.operator.op(mri_2D)\n\n# Before training, here is the simple reconstruction we have using a\n# density compensated adjoint.\ndc_adjoint = model.operator.adj_op(kspace_mri_2D)\nfig, axs = plt.subplots(1, 3, figsize=(15, 5))\nplot_state(axs, mri_2D, init_traj, dc_adjoint)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Start training loop\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "epoch = 100\noptimizer = torch.optim.RAdam(model.parameters(), lr=1e-3)\nlosses = []  # Store the loss values and create an animation\nimage_files = []  # Store the images to create a gif\nmodel.train()\n\nwith tqdm(range(epoch), unit=\"steps\") as tqdms:\n    for i in tqdms:\n        out = model(kspace_mri_2D)  # Forward pass\n\n        loss = torch.nn.functional.l1_loss(out, mri_2D[None])  # Compute loss\n        tqdms.set_postfix({\"loss\": loss.item()})  # Update progress bar\n        losses.append(loss.item())  # Store loss value\n\n        optimizer.zero_grad()  # Zero gradients\n        loss.backward()  # Backward pass\n        optimizer.step()  # Update weights\n\n        # Generate images for gif\n        hashed = joblib.hash((i, \"learn_traj\", time.time()))\n        filename = f\"{tmp.NamedTemporaryFile().name}.png\"\n        fig, axs = plt.subplots(2, 2, figsize=(10, 10))\n        plot_state(\n            axs,\n            mri_2D,\n            init_traj,\n            out,\n            losses,\n            save_name=filename,\n        )\n        image_files.append(filename)\n\n\n# Make a GIF of all images.\nimgs = [Image.open(img) for img in image_files]\nimgs[0].save(\n    \"mrinufft_learn_unet.gif\",\n    save_all=True,\n    append_images=imgs[1:],\n    optimize=False,\n    duration=2,\n    loop=0,\n)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "Reconstruction from partially trained U-Net model\n\n"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": null,
      "metadata": {
        "collapsed": false
      },
      "outputs": [],
      "source": [
        "model.eval()\nnew_recon = model(kspace_mri_2D)\nfig, axs = plt.subplots(2, 2, figsize=(10, 10))\nplot_state(axs, mri_2D, init_traj, new_recon, losses)\nplt.show()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "## References\n\n.. [fastmri] O. Ronneberger, P. Fischer, and Thomas Brox. U-net: Convolutional networks\n          for biomedical image segmentation. In International Conference on Medical\n          image computing and computer-assisted intervention, pages 234\u2013241.\n          Springer, 2015.\n          https://github.com/facebookresearch/fastMRI/blob/main/fastmri/models/unet.py\n\n"
      ]
    }
  ],
  "metadata": {
    "kernelspec": {
      "display_name": "Python 3",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.10.15"
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}